import pytest

from cuenca import CardTransaction


@pytest.mark.vcr
def test_card_transaction_parent_retrieve():
    ct = CardTransaction.retrieve('CT001')
    related_cts = ct.related_card_transactions
    assert len(related_cts) == 3
    for card_transaction in related_cts:
        assert card_transaction.type != 'auth'


@pytest.mark.vcr
def test_card_transaction_child_retrieve():
    ct = CardTransaction.retrieve('CT002')
    related_cts = ct.related_card_transactions
    assert len(related_cts) == 1
    assert related_cts[0].type == 'auth'


@pytest.mark.vcr
def test_card_transaction_refund_without_parent_id_retrieve():
    ct = CardTransaction.retrieve('CT003')
    assert not ct.related_card_transactions


@pytest.mark.vcr
def test_card_transaction_expiration_retrieve():
    ct = CardTransaction.retrieve('CT004')
    assert ct.type == 'expiration'



import datetime as dt
import re
from typing import ClassVar, List, Optional, Tuple, cast
from urllib.parse import quote

from cuenca_validations.types import ApiKeyQuery, ApiKeyUpdateRequest
from pydantic.dataclasses import dataclass

from ..http import Session, session as global_session
from .base import Creatable, Queryable, Retrievable, Updateable


@dataclass
class ApiKey(Creatable, Queryable, Retrievable, Updateable):
    _resource: ClassVar = 'api_keys'
    _query_params: ClassVar = ApiKeyQuery

    secret: str
    deactivated_at: Optional[dt.datetime]
    user_id: Optional[str]
    metadata: Optional[str]

    @property
    def active(self) -> bool:
        return (
            self.deactivated_at is None
            or self.deactivated_at > dt.datetime.utcnow()
        )

    @classmethod
    def create(cls, *, session: Session = global_session) -> 'ApiKey':
        return cast('ApiKey', cls._create(session=session))

    @classmethod
    def deactivate(cls, api_key_id: str, minutes: int = 0) -> 'ApiKey':
        """
        deactivate an ApiKey in a certain number of minutes. If minutes is
        negative, the API will treat it the same as 0. You can't deactivate
        the same key with which the client is configured, since that'd risk
        locking you out. The deactivated key is returned so that you have the
        exact deactivated_at time.
        """
        url = cls._resource + f'/{api_key_id}'
        resp = global_session.delete(url, dict(minutes=minutes))
        return cast('ApiKey', cls._from_dict(resp))

    @classmethod
    def update(
        cls,
        api_key_id: str,
        metadata: Optional[str] = None,
        user_id: Optional[str] = None,
    ) -> 'ApiKey':
        """
        If the current user has enough permissions, it associates an ApiKey to
        the `user_id` or updates the correspoding metadata
        """
        req = ApiKeyUpdateRequest(metadata=metadata, user_id=user_id)
        resp = cls._update(api_key_id, **req.dict(exclude_none=True))
        return cast('ApiKey', resp)

    @classmethod
    def validate(
        cls, permissions: List[str], *, session: Session = global_session
    ) -> Tuple[Optional[str], List[str]]:
        """
        User this method to validate if your credentials have access to a set
        of permissions.

        Parameters:
        permissions (List[str]): The cuenca URL's to be validated.
        Ej. cuenca://oaxaca/{user_id}/transfers.read

        Returns:
        str: The user_id associated to the required permissions, None if it
              applies to everyone
        List[str]: The subset of `permissions` approved, an empty list if
        nothing was approved
        """
        quoted = quote(','.join(permissions))
        resp = session.get('/authorizations', dict(actions=quoted))

        authorized_permissions = []
        user_id = None
        for permission in permissions:
            # The response could have a user_id or '*' if it applies to
            # everyone. Convert the permission to a reg exp to find the user
            pattern = re.escape(permission)
            pattern = permission.format(user_id=r'(?P<user_id>US[\w=-]+|\*)')

            # Get the first match in the allowed actions, continue if not
            for allowed_permission in resp['allow']:
                match = re.match(pattern, allowed_permission)
                if match:
                    break
            else:
                continue

            # If there is a `user_id` group, assigned it
            try:
                match_user_id = match.group('user_id')
            except IndexError:
                ...
            else:
                user_id = match_user_id if match_user_id != '*' else None
            authorized_permissions.append(permission)

        return user_id, authorized_permissions



import datetime as dt
from unittest.mock import patch
from urllib.parse import quote

import pytest

from cuenca import ApiKey


@pytest.mark.vcr
def test_api_keys_create():
    api_key = ApiKey.create()
    assert api_key.id is not None
    assert api_key.secret is not None
    assert api_key.active


@pytest.mark.vcr
def test_api_keys_retrieve():
    id_key = 'AKe2V4KrZUQ7-GTmzvjexvrw'
    api_key: ApiKey = ApiKey.retrieve(id_key)
    assert api_key.id == id_key
    assert api_key.secret == '********'


@pytest.mark.vcr
def test_api_key_deactivate():
    id_key = 'AKSMPzsvVPROyLZhkwcOtHxA'
    api_key: ApiKey = ApiKey.retrieve(id_key)
    assert api_key.active

    disabled = ApiKey.deactivate(id_key, 0)
    assert disabled.id == api_key.id
    assert disabled.deactivated_at is not None
    assert not disabled.active

    api_key.refresh()
    assert not api_key.active


@pytest.mark.vcr
def test_update_api_key():
    fields_to_update = dict(
        metadata='ANYTHING', user_id='USiBeLDwEWT_inkyE4CrRsrQ'
    )
    api_key_id = 'AKkVaALThyQ3SSFeR-qBAKiw'

    updated = ApiKey.update(api_key_id, **fields_to_update)
    assert all(
        getattr(updated, key) == value
        for key, value in fields_to_update.items()
    )

    api_key = ApiKey.retrieve(api_key_id)
    assert all(
        getattr(api_key, key) == value
        for key, value in fields_to_update.items()
    )


def test_api_key_to_dict():
    created = dt.datetime.now()
    date = created.astimezone(dt.timezone.utc).isoformat()
    api_key: ApiKey = ApiKey(
        id='12345',
        secret='********',
        created_at=created,
        deactivated_at=None,
        updated_at=date,
        metadata=None,
        user_id=None,
    )
    api_key_dict = dict(
        id='12345',
        secret='********',
        created_at=date,
        deactivated_at=None,
        updated_at=date,
        metadata=None,
        user_id=None,
    )
    assert api_key_dict == api_key.to_dict()


def test_api_key_from_dict():
    api_keys_dict = dict(
        id='123455',
        secret='*********',
        created_at=dt.datetime.utcnow(),
        deactivated_at=None,
        extra_field_1='not necessary',
        extra_field_2=12345,
        updated_at=dt.datetime.utcnow(),
        metadata=None,
        user_id=None,
    )
    api_key = ApiKey._from_dict(api_keys_dict)
    assert not hasattr(api_key, 'extra_field_1')
    assert not hasattr(api_key, 'extra_field_2')
    assert api_key.id is not None


@pytest.mark.parametrize(
    'permissions, response, result',
    [
        (['cuenca://oaxaca/{user_id}/transfers.read]'], [], (None, [])),
        (
            ['cuenca://oaxaca/{user_id}/transfers.read'],
            ['cuenca://no_oaxaca/*/transfers.read'],
            (None, []),
        ),
        (
            [
                'cuenca://oaxaca/{user_id}/transfers.read',
                'cuenca://oaxaca/{user_id}/transfers.write',
            ],
            ['cuenca://oaxaca/US12345/transfers.read'],
            ('US12345', ['cuenca://oaxaca/{user_id}/transfers.read']),
        ),
        (
            [
                'cuenca://oaxaca/{user_id}/transfers.read',
                'cuenca://oaxaca/{user_id}/transfers.write',
            ],
            ['cuenca://oaxaca/*/transfers.read'],
            (None, ['cuenca://oaxaca/{user_id}/transfers.read']),
        ),
        (
            ['cuenca://oaxaca/transfers.read'],
            ['cuenca://oaxaca/transfers.read'],
            (None, ['cuenca://oaxaca/transfers.read']),
        ),
        (
            ['cuenca://oaxaca/transfers.read'],
            ['cuenca://oaxaca/transfers.write'],
            (None, []),
        ),
    ],
)
@patch('cuenca.http.session.get')
def test_validate_not_approved(mocked_get, permissions, response, result):
    mocked_get.return_value = dict(allow=response)
    assert result == ApiKey.validate(permissions)
    quoted = quote(','.join(permissions))
    mocked_get.assert_called_once_with('/authorizations', dict(actions=quoted))
